#!/usr/bin/perl -w

use strict;
use CGI;
use Getopt::Long;
use HTML::Parser;
use LWP::UserAgent;
use URI::URL;

use vars qw($program $version);
$program = 'Html2Wml';
$version = '0.3.0b5';

=head1 NAME

Html2Wml - Program that can convert HTML pages to WML pages

=head1 SYNOPSIS

in a shell:

    html2wml.cgi [--[no]collapse] [--[no]linearize] [--[no]xmlcheck] <file|url>

as a CGI:

    /cgi-bin/html2wml.cgi?url=<url>

=head1 DESCRIPTION

Html2Wml tries to convert (quite loosely) HTML pages to WML pages. 
WML support is currently minimal: the result pages should be viewable on 
Wap browser, but there are not optimized. 

Please note that currently, the output depends on the quality of the 
input pages. 

=head1 OPTIONS

=over 4

=item collapse

This option tells Html2Wml to collapse redundant white space characters 
and empty paragraphs. This option is on by default, but you can desctivate 
this by using --nocollapse. 

This behavior is not really standard, but the aim is reducing the size of 
the output. WML pages are primarily intented for Wap devices, which usually 
have slow connections. The smaller the WML result is, the faster it can be 
downloaded. Furthermore, collapsing white spaces is the normal behavior 
for HTML pages. 

Empty paragraphs are also collapsed (this is really not standard), but it 
should avoid empty screens: the display of a Wap device is usually small, 
and it can be annoying to scroll down a lot because of empty lines. 

=item linearize

This options is on by default. It makes Html2Wml linearize the table 
I<E<agrave> la> Lynx. I think it is better than trying to use WML tables 
because, contrary to HTML tables, they have extremely limited features, 
and it could be a waste of time. 

=item xmlcheck

When this option is on, it send the WML output to XML::Parser to check 
its well-formedness. 

=back

=head1 VERSION HISTORY

=over 4

=item v0.3.0 beta 5, Thurdsay, December 6, 2000

added a collapse option to control whether redundant white space chars 
are kept or not

added the documentation for the options

=item v0.3.0 beta 4, Monday, December 4, 2000

added a linearize option (on by default) so that it linearizes tables

normal texts (i.e. not inside <pre> tags) are cleaned by stripping 
multiples spaces and lines

the syntax repair engine has been rewritten, but it does not work in 
all cases

=item v0.3.0 beta 3, Sunday, November 19, 2000

first try to add syntax check/repair

=item v0.2.3, Saturday, November 18, 2000

added URI translation

=item v0.2.2, Thursday, November 15, 2000

added XML checker; the output is given to Expat to check that it is 
well-formed XML. Note that XML::Parser is loaded on demand.

added options support (using Getopt::Long)

=item v0.2.1, Tuesday, November 14, 20000

cleaned up a few things;

added some attributes to the allow list

added url resolution

=item v0.2.0, Saturday, November 11, 2000

first functional version; now uses HTML::Parser v3

=item v0.1.0, Friday, November 10, 2000

prototype version

=back

=head1 AUTHOR

SE<eacute>bastien Aperghis-Tramoni <madingue@resus.univ-mrs.fr>

=head1 COPYRIGHT

Html2Wml is Copyright (C)2000 SE<eacute>bastien Aperghis-Tramoni.

This program is free software. You can redistribute it and/or modify it 
under the terms of either the Perl Artistic License or the GNU General 
Public License, version 2 or later. 

=cut


# 
# globals
# 
use vars qw($cgi);
$cgi = 0;
my $result;
my %options = (
    help       => 0,    ## show the usage and exit
    version    => 0,    ## show the program name and version and exit
    
    collapse   => 1,    ## when on, collapse white space characters
    linearize  => 1,    ## when on, suppress the tables tags
    xmlcheck   => 0,    ## when on, perform a well-formedness check (using XML::Parser)
);

## used by the html parser
use vars qw(%state);
%state = (
    doc_uri => '',  ## document absolute URI
    output  => '',  ## buffer for storing output
    skip    => 0,   ## skip switch (on/off)
    stack   => []   ## tag stack
);

my %dtdent = ();
    $dtdent{emph}   = 'em,strong,b,i,u,big,small';
    $dtdent{layout} = 'br';
    $dtdent{text}   = $dtdent{emph};
    $dtdent{flow}   = "$dtdent{text},$dtdent{layout},img,anchor,a,table";
    $dtdent{fields} = "$dtdent{flow},input,select,fieldset";

my %with = (
    html    => { action => 'replace',  new_value => 'wml'  }, 
    wml     => { action => 'keep',     nest => 'head,card'  }, 
    
    ## header tags
   # meta    => { action => 'keep',     nest => 'EMPTY' }, 
    title   => { action => 'skip' }, 
    style   => { action => 'skip' }, 
    script  => { action => 'skip' }, 
    
    ## structural tags
    body    => { action => 'replace',  new_value => 'card' }, 
    card    => { action => 'keep',     nest => 'do,p,pre' }, 
    h1      => { action => 'replace',  new_value => 'p' }, 
    h2      => { action => 'replace',  new_value => 'p' }, 
    h3      => { action => 'replace',  new_value => 'p' }, 
    li      => { action => 'replace',  new_value => 'p' }, 
    dt      => { action => 'replace',  new_value => 'p' }, 
    dd      => { action => 'replace',  new_value => 'p' }, 
    div     => { action => 'replace',  new_value => 'p' }, 
    p       => { action => 'keep',     nest => "$dtdent{fields},do" }, 
    br      => { action => 'keep',     nest => 'EMPTY' }, 
    pre     => { action => 'keep',     nest => 'a,br,i,b,em,strong,input,select' }, 
    tt      => { action => 'replace',  new_value => 'pre' }, 
    
    ## tables tags
    table   => { action => 'keep',     nest => 'tr',  attributes => 'colspan,rowspan' }, 
    caption => { action => 'skip' }, 
   'tr'     => { action => 'keep',     nest => 'td' }, 
    th      => { action => 'replace',  new_value => 'td' }, 
    td      => { action => 'keep',     nest => "$dtdent{emph},$dtdent{layout},img,a,anchor" }, 
    
    ## link tags
    a       => { action => 'keep',     nest => 'br,img',  attributes => 'href,name' }, 
    img     => { action => 'keep',     nest => 'EMPTY',  attributes => 'src,alt,align' }, 
    
    ## form tags -- currently not handled
   # select  => { action => 'keep',     nest => 'optgroup,option' }, 
   # input   => { action => 'keep',     nest => 'EMPTY' }, 
    
    ## font tags
    em      => { action => 'keep',     nest => $dtdent{flow} }, 
    strong  => { action => 'keep',     nest => $dtdent{flow} }, 
    b       => { action => 'keep',     nest => $dtdent{flow} }, 
    i       => { action => 'keep',     nest => $dtdent{flow} }, 
    u       => { action => 'keep',     nest => $dtdent{flow} }, 
    big     => { action => 'keep',     nest => $dtdent{flow} }, 
    small   => { action => 'keep',     nest => $dtdent{flow} }
);


# 
# main
# 
if(@ARGV) {
    ## launched from shell
    
    ## getting options
    GetOptions(\%options, map { "$_!" } keys %options);
    version() if $options{version};
    usage() if $options{help};
    usage() unless @ARGV;
    apply_options();
    
    ## converting the file
    $result = html2wml(shift);
    
} else {
    ## launched from web
    $cgi = new CGI;
    
    ## get the options
    for my $option (keys %options) {
        $options{$option} = $cgi->param($option)
    }
    
    apply_options();
    
    $result = html2wml($cgi->param('url') or cgiusage());
}

## XML check
if($options{xmlcheck}) {
    require XML::Parser;
    my $xmlparser = new XML::Parser Style => 'Tree', ErrorContext => 2;
    $xmlparser->parse($result);
}


print $cgi->header(-status => 'text/vnd.wap.wml') if $cgi;
print $result;



# 
# apply_options()
# -------------
sub apply_options {
    if($options{linearize}) {
        delete @with{qw(table tr td th)};
        $with{'tr'} = { action => 'replace', new_value => 'p' };
    }
}

# 
# html2wml()
# --------
sub html2wml {
    my $url = shift;
    my $page;
    my %attr;
    my $converter = new HTML::Parser api_version => 3;
    my $date = localtime;
    
    ## read the file 
    if($url =~ m{http://}) {  ## absolute uri
        $page = get_url($url)
    
    } elsif(-f $url) {        ## local file
        $page = read_file($url)
    
    } else {                  ## absolute url relative to the server
        $page = get_url( URI::URL->new($url, $cgi->self_url)->abs )
    }
    
    ## Gather information from the HTML page
    if($page =~ m/<title>(.*)<\/title>/is) {
        $attr{title} = clean_spaces($1)
    }
    
    ## conversion
    $state{skip} = 0;
    $state{output} = q|<?xml version="1.0"?>| . $/ 
                   . q|<!DOCTYPE wml PUBLIC "-//WAPFORUM//DTD WML 1.2//EN" |
                   . q|"http://www.wapforum.org/DTD/wml12.dtd">| . $/ 
                   .qq|<!-- Converted by $program $version on $date -->| . $/;
    
    $converter->unbroken_text(1);
    $converter->handler(start       => \&start,  'tagname, attr, attrseq, text');
    $converter->handler(end         => \&end,    'tagname, text');
    $converter->handler(text        => \&text,   'text, is_cdata');
    $converter->handler(declaration => '');
    $converter->handler(comment     => '');
    $converter->handler(process     => '');
    
    $converter->parse($page);
    $converter->eof;
    
    ## post-conversion clean-up
    $state{output} =~ s/\n+/\n/go  if $options{collapse};
    $state{output} =~ s/\s+>/>/go;
    $state{output} =~ s|\s+/>|/>|go;
    
    if(length $attr{title}) {
        $state{output} =~ s/<card/<card title="$attr{title}"/;
    }
    
    return $state{output}
}


# 
# get_url()
# -------
# This function gets and returns the file from the given URI. 
# 
sub get_url {
    my $uri = shift;
    my $ua = new LWP::UserAgent;
    $ua->agent("$program/$version " . $ua->agent);
    my $request = new HTTP::Request GET => $uri;
    my $response = $ua->request($request);
    
    if($response->is_error) {
        return $response->error_as_HTML
    }
    
    $state{doc_uri} = $uri;
    return $response->content
}


# 
# read_file()
# ---------
# This function reads and returns the file from the local disk. 
# 
sub read_file {
    my $file = shift;
    local $/ = undef;
    open(FILE, $file) or die "can't read file '$file': $! ";
    $file = <FILE>;
    close(FILE);
    return $file
}


# 
# clean_spaces()
# ------------
sub clean_spaces {
    my $str = shift;
    $str =~ s/ +/ /go;
    $str =~ s/^ //;
    $str =~ s/ $//;
    return $str
}


# 
# Functions used by the HTML parser
#  - start() is called on a start tag event
#  - end() is called on an end tag event
#  - text() is called on a text event
# 
  
sub start {
    my($tag, $attr) = @_;
    
    ## reconstruct well-formed attributes list with only the allowed ones
    $attr = join ' ', map { $_.'="' . (/href|src/ ? xlate_url($attr->{$_}) : $attr->{$_}) . '"' if $attr->{$_} } 
                      split(',', $with{$tag}{attributes});
    $attr = ' ' . $attr if length $attr;
    
    ## set the skip mode state
    $state{skip} = 1 if $with{$tag}{action} eq 'skip';
    
    ## syntax check/repair
    if(@{$state{stack}} and $with{$tag}{action} =~ /keep|replace/ and $with{$tag}{nest} ne 'EMPTY') {
        my $curr_tag = ($with{$tag}{action} eq 'replace' ? $with{$tag}{new_value} : $tag);
        
        while(my $prev_tag = pop @{$state{stack}}) {
            if($with{$prev_tag}{nest} =~ /\b$curr_tag\b/) {
                push @{$state{stack}}, $prev_tag;
                last
            }
            $state{output} .= "</$prev_tag>";
        }
    }
    
    ## simple tag translation
    if($with{$tag}{action} eq 'keep') {
        if($with{$tag}{nest} eq 'EMPTY') {
            $state{output} .= "<$tag$attr/>"
        } else {
            $state{output} .= "<$tag$attr>";
            push @{$state{stack}}, $tag;
        }
    
    } elsif($with{$tag}{action} eq 'replace') {
        $state{output} .= "<$with{$tag}{new_value}$attr>";
        push @{$state{stack}}, $with{$tag}{new_value};
    
    } else {
        ## do nothing
    }
}
  
sub end {
    my($tag) = @_;
    
    $state{skip} = 0 if $with{$tag}{action} eq 'skip';
    return if $with{$tag}{nest} eq 'EMPTY';
    
    if($with{$tag}{action} eq 'keep') {
        $state{output} .= "</$tag>";
        pop @{$state{stack}};
    
    } elsif($with{$tag}{action} eq 'replace') {
        $state{output} .= "</$with{$tag}{new_value}>";
        pop @{$state{stack}};
    
    } else {
        ## do nothing
    }
}
  
sub text {
    my($text) = @_;
    $text = clean_spaces($text) if $options{collapse} and ${$state{stack}}[$#{$state{stack}}] ne 'pre';
    $state{output} .= $text unless $state{skip}
}


# 
# xlate_url()
# ---------
# This function translates the given url so that the pointed document will 
# pass throught this CGI for conversion. 
# 
sub xlate_url {
    my $url = shift;
   # my $base = URI::URL->new($cgi->param('url'), $cgi->self_url)->abs;
    
    # $url is the url from a href or a src attribute
    # $base is the _absolute_uri_ of the current document
    
    return URI::URL->new($url, $state{doc_uri})->abs
}


# 
# version()
# -------
sub version {
    print "$program v$version\n"; exit
}


# 
# usage()
# -----
sub usage {
    print <<"USAGE"; exit
usage: $program [options] file

options: 
  --xmlcheck        activate the XML check: output is given to Expat
  --noxmlcheck      desactivate the XML checker (default)
  
  --help            show this help screen and exit
  --version         show the program name and version and exit
USAGE
}


# 
# cgiusage()
# --------
sub cgiusage {
    print $cgi->header, <<"USAGE"; exit
<html>
<head>
<title>Error</title>
</head>
<body>
<h1>Error</h1>
<p>This CGI was called with incorrect parameters. 
Check your request and try again </p>
</body>
</html>
USAGE
}


